//
//  HTTPThreadMgr.cpp
//
//  This class manages the CHTTPThread classes.
//  Used to make order of outstanding requests so that
//  things do not get chaotic.
//
//  Author: Don Metzler for WDJ
//

#define STRICT

#include "stdafx.h"

#include <afxtempl.h>

#include "HThrdMgr.h"
#include "HThrd.h"

CHTTPThreadMgr::CHTTPThreadMgr()
{
    int		i ;

    // Create the threads that do the work
    for(i=0;i<MAX_THREADS;i++)
        m_pIThread[i] = new CHTTPThread() ;

    // Create the pending request CList
    m_pPendingList = new CList<PENDINGLIST,PENDINGLIST>
        (sizeof(PENDINGLIST)) ;
}

CHTTPThreadMgr::~CHTTPThreadMgr()
{
    int		i ;

    // Delete all threads
    for(i=0;i<MAX_THREADS;i++)
        delete m_pIThread[i] ;

    // Remove all pending requests from the list
    RemovePendingRequests() ;

    // Delete the pending list
    if (m_pPendingList)
        delete m_pPendingList ;
}

void CHTTPThreadMgr::OnThreadComplete(int iThreadNum, 
                                      char *szLocalFile, 
                                      int iStatus)
{
    // This notification comes in when a thread has completed
    // processing...

    if (m_pWndCallback)
    {
        WPARAM	wParam ;
        LPARAM	lParam ;

        wParam = MAKEWPARAM(iThreadNum, iStatus) ;
        lParam = (LPARAM)(LPSTR) szLocalFile ;
        m_pWndCallback->SendMessage(WM_THREADCOMPLETE, 
                                    wParam, 
                                    lParam) ;
    }

    // Check for any requests which are waiting on the list...
    CheckNextRequest() ;
}

void CHTTPThreadMgr::InitThreads(CWnd *pWndCallback, 
                                 char *szTitle)
{
    m_pWndCallback = pWndCallback ;

    int     i ;
    BOOL    bUseStatusDlg = TRUE ;

    for(i=0;i<MAX_THREADS;i++)
        if (m_pIThread[i])
            m_pIThread[i]->Init(this, i, szTitle) ;
}

void CHTTPThreadMgr::ResetAllThreads()
{
    int     i ;

    // Reset all threads
    for(i=0;i<MAX_THREADS;i++)
        if (m_pIThread[i])
            m_pIThread[i]->ResetSession() ;
}

void CHTTPThreadMgr::RemovePendingRequests()
{
    ASSERT(m_pPendingList) ;

    m_pPendingList->RemoveAll() ;
}

int CHTTPThreadMgr::GetPendingRequestCount()
{
    if (!m_pPendingList)
        return (0) ;

    if (m_pPendingList->IsEmpty())
        return (0) ;

    // Return the number of pending requests in the list
    return (m_pPendingList->GetCount()) ;
}

int CHTTPThreadMgr::GetWorkingThreadCount()
{
    int	    i, iCount ;

    iCount = 0 ;

    for(i=0;i<MAX_THREADS;i++)
    {
        if (m_pIThread[i])
            if (!m_pIThread[i]->IsThreadAvailable())
                iCount++ ;
    }

    // Return a count of all threads which are currently working
    return (iCount) ;
}

void CHTTPThreadMgr::GetRemoteFile(char *szURL, 
                                   char *szLocalFile)
{
    int     i ;

    // Is there a thread available?
    i = GetNextAvailableThread() ;

    if (i == -1)
    {
        DebugMsg("CHTTPThreadMgr - all threads are busy") ;

        // All threads are busy, so put this one in the list...
        AddToPendingList(szURL, szLocalFile) ;
    }
    else
    {
        DebugMsg("CHTTPThreadMgr - requesting file") ;

        // Go ahead and start up the thread to get the file...
        RequestFile(i, szURL, szLocalFile) ;
    }
}

void CHTTPThreadMgr::AddToPendingList(char *szURL, 
                                      char *szLocalFile)
{
    PENDINGLIST	    pList ;

    ASSERT(m_pPendingList) ;

    DebugMsg("CHTTPThreadMgr::AddToPendingList()") ;

    // Add the URL and local filename to the pending list
    strcpy(pList.szURL, szURL) ;
    strcpy(pList.szLocalFile, szLocalFile) ;

    // Get the time we are adding to the list...
    GetLocalTime(&pList.st) ;

    // Add to the request list...
    m_pPendingList->AddTail(pList) ;
}

int CHTTPThreadMgr::GetNextAvailableThread()
{
    int	    i ;

    // Check the status of all the threads...
    for(i=0;i<MAX_THREADS;i++)
    {
        if (m_pIThread[i]->IsThreadAvailable())
            return (i) ;
    }

    // All threads are currently busy...
    return (-1) ;
}

void CHTTPThreadMgr::CheckNextRequest()
{
    // If there are any pending requests queued in the list,
    // fire off another request...

    ASSERT(m_pPendingList) ;

    if (m_pPendingList->IsEmpty())
        return ;

    int i = GetNextAvailableThread() ;

    // No threads available (this should never happen here!)
    if (i == -1)
        return ;

    PENDINGLIST     pList ;

    POSITION pos = m_pPendingList->GetHeadPosition() ;
    pList = m_pPendingList->GetAt(pos) ;

    // Have the thread start up...
    RequestFile(i, pList.szURL, pList.szLocalFile) ;

    // Remove the item from the list...
    m_pPendingList->RemoveAt(pos) ;
}

void CHTTPThreadMgr::RequestFile(int iThread, 
                                 char *szURL, 
                                 char *szLocalFile)
{
    CString	strURL = szURL ;

    DebugMsg("CHTTPThreadMgr::RequestFile()") ;

    if (m_pIThread[iThread] == NULL)
    {
        DebugMsg("CHTTPThreadMgr - thread is invalid!") ;
        return ;
    }

    // Let the thread know which number it is
    m_pIThread[iThread]->SetThreadNum(iThread) ;

    // Start the thread a-workin'...
    m_pIThread[iThread]->GetRemoteFile(strURL, szLocalFile) ;
}

void CHTTPThreadMgr::DebugMsg(char *szMsg)
{
    // Simple debugging messages using AfxMessageBox()
    AfxMessageBox(szMsg) ;
}
