// wdj.c

#include "ntddk.h"
#include "wdj.h"

typedef struct _DEVICE_EXTENSION {
    PIRP Irp;
    ULONG InLength;
    ULONG OutLength;
} DEVICE_EXTENSION, *PDEVICE_EXTENSION;

NTSTATUS WdjDispatchControl(IN PDEVICE_OBJECT, IN PIRP);
NTSTATUS WdjDispatchCreate(IN PDEVICE_OBJECT, IN PIRP);
NTSTATUS WdjDispatchClose(IN PDEVICE_OBJECT, IN PIRP);
VOID     WdjUnload(IN PDRIVER_OBJECT);

//-------------------------------------------------------------------
NTSTATUS DriverEntry(IN PDRIVER_OBJECT  DriverObject,
                     IN PUNICODE_STRING RegistryPath)
{
    NTSTATUS        status = STATUS_SUCCESS;
    PDEVICE_OBJECT  deviceObject = NULL;
    UNICODE_STRING  deviceName, dosName;

    //
    // Create an non-EXCLUSIVE device object.
    //
    RtlInitUnicodeString(&deviceName, L"\\Device\\WdjDrv");

    status = IoCreateDevice(DriverObject, sizeof(DEVICE_EXTENSION),
                            &deviceName, FILE_DEVICE_WDJDRV,
                            0, FALSE, &deviceObject);

    if (!NT_SUCCESS(status)) {
        return STATUS_UNSUCCESSFUL;
    }

    //
    // Create a symbolic link that Win32 apps can specify to gain
    // access to this driver/device
    //
    RtlInitUnicodeString(&dosName, L"\\DosDevices\\WDJDRV");

    status = IoCreateSymbolicLink(&dosName, &deviceName);
    if (!NT_SUCCESS(status)) {
        DbgPrint("WDJ.SYS: create symbolic link failed\n");
        IoDeleteDevice(deviceObject);
        return STATUS_UNSUCCESSFUL;
    }

    //
    // Create dispatch points for device control, create, close.
    //
    DriverObject->MajorFunction[IRP_MJ_CREATE]  = WdjDispatchCreate;
    DriverObject->MajorFunction[IRP_MJ_CLOSE]   = WdjDispatchClose;    
    DriverObject->MajorFunction[IRP_MJ_DEVICE_CONTROL] 
                                                = WdjDispatchControl;
    DriverObject->DriverUnload                  = WdjUnload;

    return status;

} // DriverEntry

//-------------------------------------------------------------------
NTSTATUS WdjDispatchControl(IN PDEVICE_OBJECT DeviceObject,
                            IN PIRP           Irp)
{
    NTSTATUS            status;
    PIO_STACK_LOCATION  irpStack = IoGetCurrentIrpStackLocation(Irp);
    PDEVICE_EXTENSION   deviceExt = DeviceObject->DeviceExtension;
    PUCHAR              p;
    ULONG               i;

    DbgPrint("WDJ.SYS: IRP_MJ_DEVICE_CONTROL, ");
    
    //
    // Get the pointer to the input/output buffer and it's length
    //

    switch (irpStack->Parameters.DeviceIoControl.IoControlCode) {

        case IOCTL_WDJ_REQUEST:
            DbgPrint(("IOCTL_WDJ_REQUEST\n"));

            deviceExt->Irp = Irp;
            deviceExt->InLength = 
              irpStack->Parameters.DeviceIoControl.InputBufferLength;
            deviceExt->OutLength = 
              irpStack->Parameters.DeviceIoControl.OutputBufferLength;

            //
            // Validate callers input and output buffer sizes
            //
            if ((deviceExt->InLength == 0) || 
                (deviceExt->OutLength < deviceExt->InLength)) {
                Irp->IoStatus.Status = STATUS_BUFFER_TOO_SMALL;
                break;
            }

            //
            // Mark this IRP as pending, I'll complete it later.
            //
            Irp->IoStatus.Information = 0;
            status = Irp->IoStatus.Status = STATUS_PENDING;
            IoMarkIrpPending(Irp);
            break;
 
        case IOCTL_WDJ_TRIGGER:
            DbgPrint(("IOCTL_WDJ_TRIGGER\n"));
            
            //
            // Complete this trigger IRP
            //
            status = STATUS_SUCCESS;
            Irp->IoStatus.Status = STATUS_SUCCESS;
            Irp->IoStatus.Information = 0;
            IoCompleteRequest(Irp, IO_NO_INCREMENT);

            //
            // If there's a pending IRP, complete it.
            //
            if (deviceExt->Irp == NULL) {
                break;
            }

            p = deviceExt->Irp->AssociatedIrp.SystemBuffer;
            for (i = 0; i < deviceExt->InLength; i++) {
                (*p) += 1;    // increment each byte by 1
                p++;    
            }
            
            deviceExt->Irp->IoStatus.Status = STATUS_SUCCESS;
            deviceExt->Irp->IoStatus.Information = 
                    deviceExt->OutLength;
            IoCompleteRequest(deviceExt->Irp, IO_NO_INCREMENT);
            deviceExt->Irp = NULL;            
            break;
 
        default:
            DbgPrint("Unknown IOCTL\n");
            Irp->IoStatus.Information = 0;
            status = Irp->IoStatus.Status = STATUS_NOT_SUPPORTED;
            IoCompleteRequest(Irp, IO_NO_INCREMENT);
            break;
    }

    return status;

} // WdjDispatchControl

//-------------------------------------------------------------------
NTSTATUS WdjDispatchCreate(IN PDEVICE_OBJECT DeviceObject,
                           IN PIRP           Irp)
{
    PIO_STACK_LOCATION  irpStack = IoGetCurrentIrpStackLocation(Irp);
    PDEVICE_EXTENSION   deviceExt = DeviceObject->DeviceExtension;

    DbgPrint("WDJ.SYS: IRP_MJ_CREATE\n");
    
    Irp->IoStatus.Information = 0;
    Irp->IoStatus.Status = STATUS_SUCCESS;
    IoCompleteRequest(Irp, IO_NO_INCREMENT);
    return STATUS_SUCCESS;

} // WdjDispatchCreate

//-------------------------------------------------------------------
NTSTATUS WdjDispatchClose(IN PDEVICE_OBJECT DeviceObject,
                          IN PIRP           Irp)
{
    PIO_STACK_LOCATION  irpStack = IoGetCurrentIrpStackLocation(Irp);
    PDEVICE_EXTENSION   deviceExt = DeviceObject->DeviceExtension;

    DbgPrint("WDJ.SYS: IRP_MJ_CLOSE\n");
    
    Irp->IoStatus.Information = 0;      // meaningless for close
    Irp->IoStatus.Status = STATUS_SUCCESS;
    IoCompleteRequest(Irp, IO_NO_INCREMENT);
    return STATUS_SUCCESS;

} // WdjDispatchClose


//-------------------------------------------------------------------
VOID WdjUnload(IN PDRIVER_OBJECT DriverObject)
{
    UNICODE_STRING unicodeString;

    //
    // Delete the symbolic link and the device object.
    //
    RtlInitUnicodeString(&unicodeString, L"\\DosDevices\\WDJDRV");
    IoDeleteSymbolicLink(&unicodeString);
    IoDeleteDevice(DriverObject->DeviceObject);

} // WdjUnload
